"""
Benchmark tests.

These tests are skipped by default as a relatively large dataset must be generated
first.

See the documentation on how to run these tests:
https://docs.pola.rs/development/contributing/test/#benchmark-tests
"""
